/* ffun_lib.c --- Sortiranje binarnih (neformatiranih) datoteka.

   Datoteka sadrzi sljedece funkcije:
   - funkciju za usporedjivanje objekata u datoteci,
   - osnovne funkcije za operacije s datotekom, uz kontrolu greske,
   - genericke funkcije za sortiranje, simulacijom rada na polju
     (indeksi su tipa long int, kao kod ftell, fseek).
*/

#include <stdio.h>
#include <stdlib.h>
#include "ffun_lib.h"

/* ======================================================================== */

    /* Tip objekata u datoteci je obj_type = int (definiran u ffun_lib.h),
       a velicina takvog objekta je obj_size. */

size_t obj_size = sizeof(obj_type);

/* ======================================================================== */

    /* Funkcija za usporedjivanje dva objekta. Objekti su zadani
       pokazivacima na njih, a rezultat je (kao kod strcmp):
       < 0  <=> prvi je < drugog,  = 0  <=> prvi == drugi,
       > 0  <=> prvi je > drugog. */

    /* U nasem slucaju, obj_cmp je usporedba cijelih brojeva.
       Uocite da else-ovi nisu potrebni. */

int obj_cmp(const obj_type *p_a, const obj_type *p_b)
{
    if (*p_a < *p_b) return -1;
    else if (*p_a > *p_b) return 1;
    else return 0;
}

/* ======================================================================== */

    /***** Osnovne funkcije na datoteci s kontrolom greske. *****/

    /* Postavljanje trazene pozicije u datoteci (funkcijom fseek)
       uz kontrolu greske. */

void postavi_fpos(FILE *f_obj, long int offset, int origin)
{
    if (fseek(f_obj, offset, origin)) {
        fprintf(stderr, "Greska u fseek.\n");
        exit(10);
        }
    return;
}

    /* Broj objekata u datoteci, s provjerom duljine datoteke i
       kontrolom je li duljina cjelobrojni visekratnik duljine objekta.
       U slucaju greske, funkcija prekida izvrsavanje programa.
       Izlazni tip je long int, a ne size_t, tako da za praznu datoteku
       (n_obj = 0) smijemo koristiti n_obj - 1 u istom tipu. */

long int broj_objekata(FILE *f_obj, size_t obj_size)
{
    long int duljina;
    long int n_obj;    /* Namjerno tip NIJE size_t. */

        /* Nadji duljinu datoteke --- pomakom na kraj. */
    postavi_fpos(f_obj, 0L, SEEK_END);
    duljina = ftell(f_obj);

        /* Provjeri rezultat funkcije ftell. */
    if (duljina < 0L) {
        fprintf(stderr, "Greska u ftell!\n");
        exit(11);
    }

        /* Provjera duljine - odgovara li cijelom broju objekata. */
    if (duljina % obj_size != 0L) {
        fprintf(stderr, "Greska u duljini/sadrzaju datoteke!\n");
        exit(12);
    }

        /* Broj objekata u datoteci. */
    n_obj = duljina / obj_size;

    return n_obj;
}

    /* Praznjenje/pisanje spremnika u datoteku (funkcijom fflush)
       nakon pisanja (pri prijelazu s pisanja na citanje),
       uz kontrolu greske. */

void isprazni_spremnik(FILE *f_obj)
{
    if (fflush(f_obj)) {
        fprintf(stderr, "Greska u fflush.\n");
        exit(19);
        }
    return;
}

/* ======================================================================== */

    /* Citanje jednog objekta s trenutne pozicije u datoteci. */

void ucitaj_obj(FILE *f_obj, obj_type *p_x, size_t obj_size)
{
    if (fread(p_x, obj_size, 1, f_obj) != 1)
        if (ferror(f_obj)) {
            fprintf(stderr, "Greska u citanju.\n");
            exit(20);
        }
    return;
}

    /* Pisanje jednog objekta na trenutnu poziciju u datoteci. */

void napisi_obj(FILE *f_obj, obj_type *p_x, size_t obj_size)
{
    if (fwrite(p_x, obj_size, 1, f_obj) != 1)
        if (ferror(f_obj)) {
            fprintf(stderr, "Greska u pisanju.\n");
            exit(21);
        }
    return;
}

/* ======================================================================== */

    /***** Genericke funkcije na datoteci --- simulacija polja. *****/

    /* Citanje i-tog objekta x[i] iz datoteke (i pocinje od 0).
       Indeksi su tipa long int, kao kod ftell, fseek. */

void ucitaj_obj_i(FILE *f_obj,
                  long int i, obj_type *p_x, size_t obj_size)
{
        /* Pozicija i-tog objekta u datoteci. */
    long int f_pos = i * obj_size;

    postavi_fpos(f_obj, f_pos, SEEK_SET);

    if (fread(p_x, obj_size, 1, f_obj) != 1)
        if (ferror(f_obj)) {
            fprintf(stderr, "Greska u citanju.\n");
            exit(20);
        }
    return;
}

    /* Pisanje i-tog objekta x[i] u datoteku (i pocinje od 0).
       Indeksi su tipa long int, kao kod ftell, fseek. */

void napisi_obj_i(FILE *f_obj,
                  long int i, obj_type *p_x, size_t obj_size)
{
        /* Pozicija i-tog objekta u datoteci. */
    long int f_pos = i * obj_size;

    postavi_fpos(f_obj, f_pos, SEEK_SET);

    if (fwrite(p_x, obj_size, 1, f_obj) != 1)
        if (ferror(f_obj)) {
            fprintf(stderr, "Greska u pisanju.\n");
            exit(21);
        }
    return;
}

    /* Zamjena i-tog i j-tog objekta u datoteci.
       Indeksi su tipa long int, kao kod ftell, fseek. */

void zamjena_obj_ij(FILE *f_obj,
                    long int i, long int j, size_t obj_size)
{
    obj_type x_i, x_j;

    if (i == j) return;

        /* Ucitaj i-ti objekt x[i]. */
    ucitaj_obj_i(f_obj, i, &x_i, obj_size);

        /* Ucitaj j-ti objekt [xj]. */
    ucitaj_obj_i(f_obj, j, &x_j, obj_size);

        /* Napisi x[i] na mjesto x[j]. */
    napisi_obj_i(f_obj, j, &x_i, obj_size);

        /* Napisi x[j] na mjesto x[i]. */
    napisi_obj_i(f_obj, i, &x_j, obj_size);

        /* Isprazni spremnik u datoteku nakon pisanja,
           pri prijelazu s pisanja na citanje. */
    isprazni_spremnik(f_obj);

    return;
}

/* ======================================================================== */

    /* Ispis sadrzaja datoteke objekata. */

void ispisi_dat(FILE *f_obj, size_t obj_size)
{
    obj_type x;
    size_t broj_obj = 0;

        /* Pozicioniranje na pocetak, uz brisanje indikatora. */
    rewind(f_obj);

        /* Sekvencijalno binarno citanje do kraja datoteke. */
    while (fread(&x, obj_size, 1, f_obj) == 1) {
        ++broj_obj;
        printf("   zapis %2u: %6d\n",
               broj_obj, x);
    }

    if (ferror(f_obj)) {
        fprintf(stderr, "Greska u citanju.\n");
        exit(30);
    }

    return;
}

/* ======================================================================== */

    /* Provjera uzlazne sortiranosti datoteke objekata. */

int provjeri_sort_dat(FILE *f_obj, size_t obj_size)
{
    obj_type x_1, x_2, *p_1 = &x_1, *p_2 = &x_2, *temp;

        /* Pozicioniranje na pocetak, uz brisanje indikatora. */
    rewind(f_obj);

        /* Ucitaj prvi element x[0]. */
    ucitaj_obj(f_obj, p_1, obj_size);

        /* Sekvencijalno binarno citanje do kraja datoteke. */
    while (fread(p_2, obj_size, 1, f_obj) == 1) {

            /* Pokazivac p_1 uvijek pokazuje na prosli objekt,
               a pokazivac p_2 pokazuje na novoucitani. */

        if (obj_cmp(p_1, p_2) > 0)    /* Pogresan poredak. */
            return 0;

            /* Umjesto kopiranja novog elementa u prosli,
               napravimo zamjenu pokazivaca na elemente.
               To je bitno brze kad su elementi veliki. */

        temp = p_1;
        p_1 = p_2;
        p_2 = temp;
    }

    if (ferror(f_obj)) {
        fprintf(stderr, "Greska u citanju.\n");
        exit(31);
    }

    return 1;
}

/* ======================================================================== */
